/*
    skyeye_lcd_win32.c - LCD display emulation in an GDI32 window.
    Copyright (C) 2006 Anthony Lee

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "armdefs.h"
#include "skyeye_lcd.h"

#undef WORD
#include <windows.h>

extern mem_bank_t *bank_ptr(ARMword addr);

extern unsigned int Pen_buffer[8];

typedef struct SkyEyeLCD {
	int fWidth;
	int fHeight;
	int fDepth;

	HANDLE fThread;
#ifdef __MINGW32__
	unsigned fThreadId;
#else
	ULONG fThreadId;
#endif
	HANDLE fSem;

	ATOM fAppClass;
	HDC fScreenHDC;
	HDC fHDC;
	HBITMAP fBitmap;
	HWND fWindow;

	BOOL fNeededToRedraw;
	unsigned int fInterval;
	unsigned char *fADDR;
	unsigned int fPen[8];
} SkyEyeLCD;

static SkyEyeLCD* SkyEyeLCD_new(int width, int height, int depth, unsigned char *addr);
static void SkyEyeLCD_delete(SkyEyeLCD *lcd);
static void SkyEyeLCD_UpdateFromSkyEye(SkyEyeLCD *lcd);
static BOOL SkyEyeLCD_ProcessEvent(SkyEyeLCD *lcd, MSG *msg, LRESULT *reslut);
static void SkyEyeLCD_DoUpdate(SkyEyeLCD *lcd, RECT r);

#ifndef GET_X_LPARAM
#define GET_X_LPARAM(a)	((int)((short)LOWORD(a)))
#endif

#ifndef GET_Y_LPARAM
#define GET_Y_LPARAM(a)	((int)((short)HIWORD(a)))
#endif


static void skPenEvent(int *buffer, int eventType, int stateType, int x, int y)
{
	buffer[0] = x;
	buffer[1] = y;
	buffer[2] = 0;		// dx
	buffer[3] = 0;		// dy
	buffer[4] = eventType;	// event from pen (DOWN,UP,CLICK,MOVE)
	buffer[5] = stateType;	// state of pen (DOWN,UP,ERROR)
	buffer[6] = 1;		// no of the event
	buffer[7] = 0;		// time of the event (ms) since ts_open
}


BOOL SkyEyeLCD_ProcessEvent(SkyEyeLCD *lcd, MSG *msg, LRESULT *result)
{
	BOOL handled = TRUE;
	RECT r = {0, 0, -1, -1};

	if(lcd == NULL || msg == NULL || result == NULL)
	{
		handled = FALSE;
	}
	else switch(msg->message)
	{
		case WM_CLOSE:
			*result = 0;
			break;

		case WM_PAINT:
			if(msg->wParam == 0) GetUpdateRect(msg->hwnd, &r, FALSE);
			SkyEyeLCD_DoUpdate(lcd, r);
			ValidateRect(msg->hwnd, NULL);
			*result = 0;
			break;

		case WM_MOUSEMOVE:
			{
				WaitForSingleObject(lcd->fSem, INFINITE); // Lock
				if(lcd->fPen[5] == 1) skPenEvent(lcd->fPen, 2, 1, GET_X_LPARAM(msg->lParam), GET_Y_LPARAM(msg->lParam));
				ReleaseSemaphore(lcd->fSem, 1, NULL); // Unlock
			}
			break;

		case WM_LBUTTONDOWN:
/*		case WM_MBUTTONDOWN:
		case WM_RBUTTONDOWN:*/
			{
				WaitForSingleObject(lcd->fSem, INFINITE); // Lock
				skPenEvent(lcd->fPen, 0, 1, GET_X_LPARAM(msg->lParam), GET_Y_LPARAM(msg->lParam));
				ReleaseSemaphore(lcd->fSem, 1, NULL); // Unlock
			}
			break;

		case WM_LBUTTONUP:
/*		case WM_MBUTTONUP:
		case WM_RBUTTONUP:*/
			{
				WaitForSingleObject(lcd->fSem, INFINITE); // Lock
				skPenEvent(lcd->fPen, 1, 0, GET_X_LPARAM(msg->lParam), GET_Y_LPARAM(msg->lParam));
				ReleaseSemaphore(lcd->fSem, 1, NULL); // Unlock
			}
			break;

		case WM_KEYDOWN: /* write to stdin */
			{
				BYTE keyState[256];
				DWORD keybuffer[3];
				DWORD nWrote;

				if(GetKeyboardState(keyState) == 0) break;
				UINT scanCode = (UINT)((msg->lParam >> 16) & 0xff);

				if(ToAscii(msg->wParam, scanCode, keyState, (LPWORD)keybuffer, 0) != 1) break;

				INPUT_RECORD input;
				input.EventType = KEY_EVENT;
				input.Event.KeyEvent.bKeyDown = TRUE;
				input.Event.KeyEvent.wRepeatCount = msg->lParam & 0xff;
				input.Event.KeyEvent.wVirtualKeyCode = msg->wParam;
				input.Event.KeyEvent.wVirtualScanCode = scanCode;
				input.Event.KeyEvent.uChar.AsciiChar = keybuffer[0];
				input.Event.KeyEvent.dwControlKeyState = 0;
				WriteConsoleInput(GetStdHandle(STD_INPUT_HANDLE), &input, 1, &nWrote);
			}
			break;

		default:
			handled = FALSE;
	}

	return handled;
}


static LRESULT CALLBACK _lcd_win32_WndProc_(HWND hWnd, UINT umsg, WPARAM wParam, LPARAM lParam)
{
	BOOL handled = FALSE;
	LRESULT result = 0;
	MSG winMsg;

	SkyEyeLCD *lcd = (SkyEyeLCD*)GetWindowLong(hWnd, 0);

	if(lcd != NULL)
	{
		winMsg.hwnd = hWnd;
		winMsg.message = umsg;
		winMsg.wParam = wParam;
		winMsg.lParam = lParam;
		handled = SkyEyeLCD_ProcessEvent(lcd, &winMsg, &result);
	}

	if(!handled) result = DefWindowProc(hWnd, umsg, wParam, lParam);

	return result;
}


static void CALLBACK SkyEyeLCD_Timer(LPVOID lpArgToCompletionRoutine, DWORD dwTimerLowValue, DWORD dwTimerHighValue)
{
	SkyEyeLCD *self = (SkyEyeLCD*)lpArgToCompletionRoutine;
	WaitForSingleObject(self->fSem, INFINITE); // Lock
	self->fNeededToRedraw = TRUE;
	ReleaseSemaphore(self->fSem, 1, NULL); // Unlock
}


#ifdef __MINGW32__
static unsigned _stdcall SkyEyeLCD_Task(void *arg)
#else
static DWORD WINAPI SkyEyeLCD_Task(void *arg)
#endif
{
	RECT r;
	char mode[100];
	MSG winMsg;

	SkyEyeLCD *self = (SkyEyeLCD*)arg;
	HINSTANCE hinstance = (HINSTANCE)GetModuleHandle(NULL);
	HANDLE timer = NULL;

	WNDCLASSEX wcApp;
	wcApp.lpszClassName = "skyeye_lcd_win32";
	wcApp.hInstance = hinstance;
	wcApp.lpfnWndProc = _lcd_win32_WndProc_;
	wcApp.hCursor = LoadCursor(NULL, IDC_HAND);
	wcApp.hIcon = 0;
	wcApp.hIconSm = 0;
	wcApp.lpszMenuName = 0;
	wcApp.hbrBackground = (HBRUSH)GetStockObject(WHITE_BRUSH);
	wcApp.style = CS_CLASSDC;
	wcApp.cbClsExtra = 0;
	wcApp.cbWndExtra = 4;
	wcApp.cbSize = sizeof(wcApp);

	self->fAppClass = RegisterClassEx(&wcApp);
	self->fScreenHDC = CreateDC("DISPLAY", NULL, NULL, NULL);
	self->fHDC = CreateCompatibleDC(self->fScreenHDC);
	self->fBitmap = CreateCompatibleBitmap(self->fScreenHDC, self->fWidth, self->fHeight);
	SelectObject(self->fHDC, self->fBitmap);
	Rectangle(self->fHDC, 0, 0, self->fWidth, self->fHeight);
	snprintf(mode, 100, "If you always see this,");
	TextOut(self->fHDC, 10, 10, mode, strlen(mode));
	snprintf(mode, 100, "thus means that depth %d unsupported!!!", self->fDepth);
	TextOut(self->fHDC, 10, 30, mode, strlen(mode));

	r.left = 100;
	r.top = 100;
	r.right = r.left + self->fWidth;
	r.bottom = r.top + self->fHeight;
	AdjustWindowRectEx(&r, WS_CAPTION | WS_MINIMIZEBOX | WS_SYSMENU, FALSE, WS_EX_DLGMODALFRAME);

	self->fWindow = CreateWindowEx(WS_EX_DLGMODALFRAME, MAKEINTATOM(self->fAppClass), "", WS_CAPTION | WS_MINIMIZEBOX | WS_SYSMENU,
				       r.left, r.top, r.right - r.left + 1, r.bottom - r.top + 1,
				       NULL, NULL, hinstance, NULL);

	SetWindowLong(self->fWindow, 0, (LONG)self);
	snprintf(mode, 100, "%dx%dx%d %s", self->fWidth, self->fHeight, self->fDepth, "SkyEye LCD & Touch Screen (Win32)");
	SetWindowText(self->fWindow, mode);
	ShowWindow(self->fWindow, SW_SHOWNORMAL);

	ReleaseSemaphore(self->fSem, 1, NULL);

	if(self->fInterval > 0 && self->fInterval != INFINITE)
	{
		if((timer = CreateWaitableTimer(NULL, FALSE, NULL)) != NULL)
		{
			if(!SetWaitableTimer(timer, NULL, self->fInterval, SkyEyeLCD_Timer, self, FALSE))
			{
				CloseHandle(timer);
				timer = NULL;
			}
		}
	}

	while(1)
	{
		if(timer != NULL)
		{
			GetMessage(&winMsg, NULL, 0, 0);
			DispatchMessage(&winMsg);
			if(winMsg.message == WM_QUIT) goto exit;
		}
		else
		{
			while(PeekMessage(&winMsg, NULL, 0, 0, PM_REMOVE))
			{
				DispatchMessage(&winMsg);
				if(winMsg.message == WM_QUIT) goto exit;
			}

			if(MsgWaitForMultipleObjects(1, &(self->fThread), FALSE, self->fInterval, QS_ALLEVENTS) == WAIT_TIMEOUT)
			{
				WaitForSingleObject(self->fSem, INFINITE); // Lock
				self->fNeededToRedraw = TRUE;
				ReleaseSemaphore(self->fSem, 1, NULL); // Unlock
			}
		}
	}

exit:
	if(timer != NULL)
	{
		CancelWaitableTimer(timer);
		CloseHandle(timer);
	}

	DestroyWindow(self->fWindow);
	DeleteObject(self->fBitmap);
	DeleteDC(self->fHDC);
	DeleteDC(self->fScreenHDC);
	UnregisterClass(MAKEINTATOM(self->fAppClass), hinstance);

	return 0;
}


SkyEyeLCD* SkyEyeLCD_new(int width, int height, int depth, unsigned char *addr)
{
	SkyEyeLCD *lcd;

	if(width <= 0 || height <= 0 || addr == NULL) return NULL;
	if((lcd = (SkyEyeLCD*)malloc(sizeof(SkyEyeLCD))) == NULL) return NULL;

	lcd->fWidth = width;
	lcd->fHeight = height;
	lcd->fDepth = depth;
	lcd->fADDR = addr;
	lcd->fInterval = 200;
	lcd->fNeededToRedraw = FALSE;
	memcpy(lcd->fPen, Pen_buffer, sizeof(unsigned int) * 8);

#ifdef __MINGW32__
	lcd->fThread = (HANDLE)_beginthreadex(NULL, 0, SkyEyeLCD_Task, lcd, CREATE_SUSPENDED, &(lcd->fThreadId));
#else
	lcd->fThread = CreateThread(NULL, 0, SkyEyeLCD_Task, lcd, CREATE_SUSPENDED, &(lcd->fThreadId));
#endif
	lcd->fSem = CreateSemaphore(NULL, 0, 1, NULL);
	ResumeThread(lcd->fThread);

	WaitForSingleObject(lcd->fSem, INFINITE);
	ReleaseSemaphore(lcd->fSem, 1, NULL);

	return lcd;
}


void SkyEyeLCD_delete(SkyEyeLCD *lcd)
{
	if(lcd)
	{
		PostThreadMessage(lcd->fThreadId, WM_QUIT, 0, 0);
		WaitForSingleObject(lcd->fThread, INFINITE);
		CloseHandle(lcd->fSem);
		CloseHandle(lcd->fThread);
		free(lcd);
	}
}


void SkyEyeLCD_DoUpdate(SkyEyeLCD *lcd, RECT r)
{
	HDC hdc;

	if(lcd == NULL || lcd->fBitmap == NULL || lcd->fWindow == NULL || (hdc = GetDC(lcd->fWindow)) == NULL)
	{
		fprintf(stderr, "%s: lcd == NULL || lcd->fBitmap == NULL || lcd->fWindow == NULL || (hdc = GetDC(lcd->fWindow)) == NULL\n", __FUNCTION__);
		abort();
		return;
	}

	if(r.right < r.left || r.bottom < r.top)
	{
		r.left = 0;
		r.top = 0;
		r.right = lcd->fWidth - 1;
		r.bottom = lcd->fHeight - 1;
	}

	WaitForSingleObject(lcd->fSem, INFINITE); // Lock
	BitBlt(hdc, r.left, r.top, r.right - r.left + 1, r.bottom - r.top + 1, lcd->fHDC, r.left, r.top, SRCCOPY);
	ReleaseSemaphore(lcd->fSem, 1, NULL); // Unlock

	ReleaseDC(lcd->fWindow, hdc);
}


static u32 colors4b[16] =
{
	0x000000,0x000080,0x008000,0x008080,0x800000,0x800080,0x808000,0x808080,
	0xc0c0c0,0x0000ff,0x00ff00,0x00ffff,0xff0000,0xff00ff,0xffff00,0xffffff
};


void SkyEyeLCD_UpdateFromSkyEye(SkyEyeLCD *lcd)
{
	int i, j;
	COLORREF color;

	if(lcd == NULL || lcd->fBitmap == NULL || lcd->fWindow == NULL) return;

	WaitForSingleObject(lcd->fSem, INFINITE); // Lock

	memcpy(Pen_buffer, lcd->fPen, sizeof(unsigned int) * 8);

	if(lcd->fNeededToRedraw == FALSE)
	{
		ReleaseSemaphore(lcd->fSem, 1, NULL); // Unlock
		return;
	}

	// FIXME: remind me what region should update from skyeye_lcd.c or device ...
	switch(lcd->fDepth)
	{
		case 4: /* 4-bit index colors */
			{
				const unsigned char *buf = lcd->fADDR;
				i = 0; j = 0;
				while(i < lcd->fHeight)
				{
					int c = colors4b[*buf & 0x0f];
					color = RGB((c >> 16) & 0xff, (c >> 8) & 0xff, c & 0xff);
					SetPixel(lcd->fHDC, j, i, color);

					if(++j >= lcd->fWidth) {i++; j = 0;}

					c = colors4b[(*buf & 0xf0) >> 4];
					color = RGB((c >> 16) & 0xff, (c >> 8) & 0xff, c & 0xff);
					SetPixel(lcd->fHDC, j, i, color);

					if(++j >= lcd->fWidth) {i++; j = 0;}

					buf++;
				}
			}
			break;

		case 8: /* 3-3-2 */
			for(i = 0; i < lcd->fHeight; i++)
			{
				const unsigned char *buf = lcd->fADDR;
				buf += i * lcd->fWidth;
				for(j = 0; j < lcd->fWidth; j++, buf++)
				{
#if 0
					u32 c = ((*buf & 0x000000e0) << 16) |
						 ((*buf & 0x0000001c) << 11) |
						 ((*buf & 0x00000003) << 6);
					color = RGB((c >> 16) & 0xff, (c >> 8) & 0xff, c & 0xff);
#endif
					color = RGB((*buf & 0xe0) | 0x1f,
						    ((*buf & 0x1c) << 3) | 0x1f,
						    ((*buf & 0x03) << 6) | 0x3f);
					SetPixel(lcd->fHDC, j, i, color);
				}
			}
			break;

		case 15: /* 5-5-5 */
			for(i = 0; i < lcd->fHeight; i++)
			{
				const u16 *buf = (const u16*)lcd->fADDR;
				buf += i * lcd->fWidth;
				for(j = 0; j < lcd->fWidth; j++, buf++)
				{
					color = RGB(((*buf & 0x7c00) >> 7) | 0x0007,
						    ((*buf & 0x03e0) >> 2) | 0x0007,
						    ((*buf & 0x001f) << 3) | 0x0007);
					SetPixel(lcd->fHDC, j, i, color);
				}
			}
			break;

		case 16: /* 5-6-5 */
			for(i = 0; i < lcd->fHeight; i++)
			{
				const u16 *buf = (const u16*)lcd->fADDR;
				buf += i * lcd->fWidth;
				for(j = 0; j < lcd->fWidth; j++, buf++)
				{
					color = RGB(((*buf & 0xf800) >> 8) | 0x0007,
						    ((*buf & 0x07e0) >> 3) | 0x0003,
						    ((*buf & 0x001f) << 3) | 0x0007);
					SetPixel(lcd->fHDC, j, i, color);
				}
			}
			break;

		case 32: /* x-8-8-8 */
			for(i = 0; i < lcd->fHeight; i++)
			{
				const u32 *buf = (const u32*)lcd->fADDR;
				buf += i * lcd->fWidth;
				for(j = 0; j < lcd->fWidth; j++, buf++)
				{
					color = RGB((*buf >> 16) & 0xff, (*buf >> 8) & 0xff, *buf & 0xff);
					SetPixel(lcd->fHDC, j, i, color);
				}
			}
			break;

		default: break;
	}

	lcd->fNeededToRedraw = FALSE;

	ReleaseSemaphore(lcd->fSem, 1, NULL); // Unlock

	PostMessage(lcd->fWindow, WM_PAINT, 1, 0);
	Sleep(0); // like SwitchToThread()
}


int win32_lcd_update(struct lcd_device *lcd_dev)
{
	if(lcd_dev == NULL || lcd_dev->priv == NULL) return 1;
	SkyEyeLCD_UpdateFromSkyEye((SkyEyeLCD*)(lcd_dev->priv));
	return 0;
}


int win32_lcd_open(struct lcd_device *lcd_dev)
{
	if(lcd_dev == NULL || lcd_dev->state == NULL) return 1;

	mem_bank_t *mbp;
	ARMul_State *state = (ARMul_State*)lcd_dev->state;

	lcd_dev->priv = NULL;
	lcd_dev->lcd_addr_end = lcd_dev->lcd_addr_begin + (lcd_dev->width * lcd_dev->height * lcd_dev->depth / 8) - 1;

	if((mbp = bank_ptr(lcd_dev->lcd_addr_begin)) == NULL)
	{
		fprintf(stderr, "No bank at address 0x%x", lcd_dev->lcd_addr_begin);
		return 1;
	}

	unsigned char *fbSkyeyeADDR = (unsigned char*)(&state->
					mem.rom[(mbp - state->mem_bank->mem_banks)][(lcd_dev->lcd_addr_begin - mbp->addr) / 4]);

	if((lcd_dev->priv = (void*)SkyEyeLCD_new(lcd_dev->width, lcd_dev->height, lcd_dev->depth, fbSkyeyeADDR)) == NULL) return 1;
	return 0;
}


int win32_lcd_close(struct lcd_device *lcd_dev)
{
	if(lcd_dev == NULL || lcd_dev->priv == NULL) return 1;
	SkyEyeLCD_delete((SkyEyeLCD*)(lcd_dev->priv));
	lcd_dev->priv = NULL;
	return 0;
}


int win32_lcd_read(struct lcd_device *lcd_dev)
{
	return 0;
}


int win32_lcd_write(struct lcd_device *lcd_dev)
{
	return 0;
}

