/*
    skyeye_win32_stdio.c - like select/read() to wait/read standard input/output/error.
    Copyright (C) 2006 Anthony Lee

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <stdio.h>
#include <windows.h>

#undef fd_set
#undef FD_ZERO
#undef FD_SET
#undef FD_ISSET
#undef FD_CLR
#define fd_set			unsigned int
#define FD_ZERO(set)		(*set = 0)
#define FD_SET(s, set)		(*set |= (1 << s))
#define FD_ISSET(s, set)	(*set & (1 << s))
#define FD_CLR(s, set)		(*set &= ~(1 << s))


int skyeye_win32_std_select(int nfds, fd_set *readfds, fd_set *writefds, fd_set *exceptfds, struct timeval *timeout)
{
	int ifds;
	int retVal = -1;
	HANDLE readHandle = NULL;
	LARGE_INTEGER due;
	HANDLE timer = NULL;

	if(nfds <= 0 || readfds == NULL || writefds != NULL || exceptfds != NULL) return -1;

	/* here we just support single pipe 0,1,2 (stdin,stdout,stderr) */
	for(ifds = 0; ifds < nfds; ifds++)
	{
		if(FD_ISSET(ifds, readfds)) break;
	}
	if(ifds > 2 || nfds - ifds > 1) return -1;

	if(timeout != NULL)
	{
		if(!(timeout->tv_sec == 0 && timeout->tv_usec == 0))
		{
			due.QuadPart = -((__int64)timeout->tv_sec * (__int64)10000000 + (__int64)timeout->tv_usec * (__int64)10);

			if((timer = CreateWaitableTimer(NULL, TRUE, NULL)) == NULL) return -1;
			if(!SetWaitableTimer(timer, &due, 0, NULL, NULL, 0))
			{
				CloseHandle(timer);
				return -1;
			}
		}
		else
		{
			due.QuadPart = 0;
		}
	}
	else
	{
		due.QuadPart = -1;
	}

	readHandle = GetStdHandle((ifds == 0 ? STD_INPUT_HANDLE : (ifds == 1 ? STD_OUTPUT_HANDLE : STD_ERROR_HANDLE)));
	if(readHandle == NULL || readHandle == INVALID_HANDLE_VALUE)
	{
		fprintf(stderr, "%s: Unable to get handle.\n", __FUNCTION__);
		if(timer != NULL) CloseHandle(timer);
		return -1;
	}


	DWORD status;
	if(timer == NULL)
	{
		status = WaitForSingleObject(readHandle, (due.QuadPart == 0 ? 0L : INFINITE));
	}
	else
	{
		HANDLE handles[2];
		handles[0] = readHandle;
		handles[1] = timer;
		status = WaitForMultipleObjects(2, handles, FALSE, INFINITE);
	}

	if(status == WAIT_TIMEOUT || (timer != NULL && status - WAIT_OBJECT_0 == 1))
	{
/*
		fprintf(stderr, "%s: TimeOut.\n", __FUNCTION__);
*/
		retVal = 0;
	}
	else if(status == WAIT_OBJECT_0)
	{
/*
		fprintf(stdout, "%s: Got it(%d).\n", __FUNCTION__, ifds);
*/
		retVal = ifds + 1;
	}
/*
	else
	{
		fprintf(stderr, "%s: Error.\n", __FUNCTION__);
	}
*/

	if(timer != NULL) CloseHandle(timer);

	return retVal;
}


ssize_t skyeye_win32_std_read(int fd, void *buf, size_t count)
{
	HANDLE handle = NULL;

	/* here we just support single pipe 0,1,2 (stdin,stdout,stderr) */
	if(fd < 0 || fd > 2 || buf == NULL || count == 0) return -1;
	handle = GetStdHandle((fd == 0 ? STD_INPUT_HANDLE : (fd == 1 ? STD_OUTPUT_HANDLE : STD_ERROR_HANDLE)));
	if(handle == NULL || handle == INVALID_HANDLE_VALUE)
	{
		fprintf(stderr, "%s: Unable to get handle.\n", __FUNCTION__);
		return -1;
	}

	DWORD nRead = 0;
	if(fd == 0) /* non-blocking stardand input */
	{
		/* FIXME: IT IS NOT GOOD */
		INPUT_RECORD input;
		DWORD nEvents = 0;
		while(!(GetNumberOfConsoleInputEvents(handle, &nEvents) == 0 || nEvents == 0) && nRead < count)
		{
			if(ReadConsoleInput(handle, &input, 1, &nEvents) == 0 || nEvents != 1) break;
			if(input.EventType != KEY_EVENT) continue;
			if(input.Event.KeyEvent.bKeyDown == FALSE) continue;
			DWORD code = input.Event.KeyEvent.wVirtualKeyCode;
			if(code == VK_SHIFT || code == VK_MENU || code == VK_CONTROL || code == VK_LWIN || code == VK_RWIN ||
			   code == VK_APPS || code == VK_CAPITAL || code == VK_SCROLL || code == VK_NUMLOCK) continue;
			*(((char*)buf) + (nRead++)) = input.Event.KeyEvent.uChar.AsciiChar;
		}
	}
	else
	{
		if(!ReadFile(handle, buf, count, &nRead, NULL)) return -1;
	}

	return (ssize_t)nRead;
}


ssize_t skyeye_win32_std_write(int fd, void *buf, size_t count)
{
	HANDLE handle = NULL;

	// here we just support single pipe 0,1,2 (stdin,stdout,stderr)
	if(fd < 0 || fd > 2 || buf == NULL || count == 0) return -1;
	handle = GetStdHandle((fd == 0 ? STD_INPUT_HANDLE : (fd == 1 ? STD_OUTPUT_HANDLE : STD_ERROR_HANDLE)));
	if(handle == NULL || handle == INVALID_HANDLE_VALUE)
	{
		fprintf(stderr, "%s: Unable to get handle.\n", __FUNCTION__);
		return -1;
	}

	DWORD nWritten = 0;
	if(!WriteFile(handle, buf, count, &nWritten, NULL)) return -1;

	return (ssize_t)nWritten;
}


#ifdef TEST_WIN32_STDIO
int main(int argc, char **argv)
{
	fd_set rset;
	struct timeval tv;
	int count = 0;
	int i = 0;
	char c;

	FD_ZERO(&rset);
	FD_SET(0, &rset);
	tv.tv_sec = 0;
	tv.tv_usec = 500000;

	SetConsoleMode(GetStdHandle(STD_INPUT_HANDLE), ENABLE_PROCESSED_INPUT);
	while(count < 10)
	{
		fprintf(stdout, "WAITTING %d ...", i); i++;
		if(skyeye_win32_std_select(1, &rset, NULL, NULL, &tv) == 1)
		{
			fprintf(stdout, "\treading");
			if(skyeye_win32_std_read(0, &c, 1) > 0)
			{
				fprintf(stdout, "\tGET: %c\n", c);
				count++;
			}
			else fprintf(stdout, "\tnone\n");
		}
		else fprintf(stdout, "\tnone\n");
	}

	return 0;
}
#endif

